import datetime
import decimal
import json
import logging

import apache_beam as beam


class LogSample(beam.PTransform):
    """Print items of the given `PCollection` to the log.

    `LogSample` prints the JSON representations of the input items
    to the Python's standard logging system.

    To avoid too much log entries being printed,
    `LogSample` limits the number of logged items.
    The constructor parameter `n` determines the limit.

    By default, `LogSample` prints logs with the `INFO` log level.
    The constructor parameter `level` determines the level.
    """

    def __init__(self, n, level=logging.INFO):
        super().__init__()
        self._n = n
        self._level = level

    def expand(self, pcoll):
        def serialize(obj):
            if isinstance(obj, (datetime.date, datetime.datetime)):
                return obj.isoformat()
            if isinstance(obj, decimal.Decimal):
                return float(obj)
            raise TypeError(f"Type {type(obj)} is not serializable")

        def log(rows):
            logger = logging.getLogger(__name__)
            for r in rows:
                logger.log(
                    self._level,
                    json.dumps(r, default=serialize, ensure_ascii=False, indent=2),
                )

        return (
            pcoll | beam.transforms.combiners.Sample.FixedSizeGlobally(self._n) | beam.ParDo(log)
        )


@beam.ptransform_fn
def ReifyMultiValueOption(pcoll, option, delimiter="|"):
    """
    Prepares multi-value delimited options

    inputs:
      - delimited string option
      - optional delimiter string (default is "|")

    output:
      - PCollection of strings

    """
    return pcoll | beam.Create([""]) | beam.FlatMap(lambda _: option.get().split(delimiter))


class MemoizedValueProviderWrapper:
    """
    Wrapper class to memoize parse result of value providers.
    Parsing function should only be called once, and result used many times

    eg. within a ParDo
    """

    def __init__(self, provider, func):
        self._value = None
        self._provider = provider
        self._func = func

    def get(self):
        if self._value is None:
            self._value = self._func(self._provider.get())
        return self._value



from copy import deepcopy

import apache_beam as beam


class _PairWithIndexNumberDoFn(beam.DoFn):
    def process(self, row):
        _, value_iter = row
        for i, value in enumerate(sorted(value_iter)):
            yield (value, i)


class PairWithIndexNumber(beam.PTransform):
    def expand(self, pcoll):
        return (
            pcoll
            | beam.Distinct()
            | beam.Map(lambda r: (None, r))
            | beam.GroupByKey()
            | beam.ParDo(_PairWithIndexNumberDoFn())
        )


@beam.ptransform_fn
def DigestCategoricalColumns(pcoll, categorical_columns):
    def to_dict(src_pcoll, column):
        dict_pcoll = (
            src_pcoll
            | f"Map[{column}]" >> beam.Map(lambda r: r[column])
            | f"Pair[{column}]" >> PairWithIndexNumber()
        )
        return beam.pvalue.AsDict(dict_pcoll)

    dicts = {cat_col: to_dict(pcoll, cat_col) for cat_col in categorical_columns}

    def set_cat_cols(row, cat_cols, **dicts):
        result = {**row}
        for cat_col in cat_cols:
            result[cat_col] = dicts[cat_col][result[cat_col]]
        return result

    return pcoll | beam.Map(set_cat_cols, categorical_columns, **dicts)


@beam.ptransform_fn
def _CreateCategoricalDict(pcoll, existing_dict_pairs):
    """
    For a specific column, creates a new "categorical dict"
    mapping values to unique ints.
    """
    existing_max_value = (
        existing_dict_pairs
        | "just values" >> beam.Map(lambda r: r[1])
        | "get max" >> beam.combiners.Top.Of(1)
        | "extract" >> beam.FlatMap(lambda r: r)
    )

    new_pairs = (
        pcoll
        | "filter for unseen"
        >> beam.Filter(
            lambda row, existing: row not in existing,
            existing=beam.pvalue.AsDict(existing_dict_pairs),
        )
        | beam.Distinct()
        | "group into single list" >> beam.combiners.ToList()
        | "append unique values"
        >> beam.FlatMap(
            lambda row, max_v: [(key, max_v + 1 + i) for i, key in enumerate(row)],
            max_v=beam.pvalue.AsSingleton(existing_max_value, default_value=0),
        )
    )

    return [new_pairs, existing_dict_pairs] | "combine new/existing" >> beam.Flatten()


@beam.ptransform_fn
def CreateCategoricalDicts(pcoll, cat_cols, existing_dict_rows):
    """
    For a set of columnular data inputs this function takes:

      - cat_cols:

          Type: [str]

          An array of "categorical" columns

      - existing_dicts:

          Type: PCollection[(string, string, int)]

          Rows of tuples of type:
            (column, previously_seen_value, mapped_unique_int)

          Mapping a set of "previously seen" keys to unique int values for each
          column.
          Not optional.
          If none exist, pass an empty PCollection.

    It then creates a transform which takes a pcollection and
      - looks at the input pcoll for unseen values in each categorical column
      - creates new unique integers for each distinct unseen value, starting at
        max(previous value for column)+1
      - ammends the existing mappings with (col, unseen_value, new_unique_int)

    Output is:

        - Type: PCollection[(string, string, int)]

    This is useful for preparing data to be trained by LightGBM
    """

    acc = []

    existing_dicts = (
        existing_dict_rows
        | beam.Map(lambda r: beam.pvalue.TaggedOutput(r[0], (r[1], r[2]))).with_outputs()
    )

    for column in cat_cols:
        existing_dict_pairs = existing_dicts[column]

        acc.append(
            pcoll
            | f"extract column value [{column}]"
            # Note: column=column used to capture value at loop time.
            #       prevents column being final value of loop due to
            #       value by reference.
            >> beam.Map(lambda r, column=column: r[column])
            | _CreateCategoricalDict(existing_dict_pairs)
            | f"re-append column [{column}]" >> beam.Map(lambda r, column=column: (column, *r))
        )

    return acc | beam.Flatten()


@beam.ptransform_fn
def ReplaceCategoricalColumns(
    inputs,
    cat_cols,
    categorical_dict_rows,
    default_unseen=None,
    features_key=None,
):
    """
    Utilizes the "categorical dictionary rows" generated by
    CreateCategoricalDicts which is a list of pairs of type of (column, value,
    unique_int), and replaces each column with the appropriate value found in
    the mapping.

    In there case where no mapping exists, `default_unseen` will be returned.
    This defaults to None, but can be set as a kwarg.
    """

    categorical_mapping_dicts = (
        categorical_dict_rows
        | "index by column" >> beam.Map(lambda r: (r[0], (r[1], r[2])))
        | "group by column" >> beam.GroupByKey()
        | "create dict per column" >> beam.MapTuple(lambda k, v: (k, dict(v)))
    )

    if features_key:
        # Nesting -- process separately from row
        def _process_nested(row, d):
            digested = {}

            features_items = row[features_key]

            for col, val in features_items.items():
                if col in cat_cols:
                    digested[col] = d.get(col, {}).get(val, default_unseen)
                else:
                    digested[col] = val

            ret = deepcopy(row)
            ret[features_key] = digested

            return ret

        return inputs | beam.Map(
            _process_nested,
            d=beam.pvalue.AsDict(categorical_mapping_dicts),
        )

    else:
        # No nesting - use row in place
        def _process(row, d):
            digested = {}

            for col, val in row.items():
                if col in cat_cols:
                    digested[col] = d.get(col, {}).get(val, default_unseen)
                else:
                    digested[col] = val

            return digested

        return inputs | beam.Map(
            _process,
            d=beam.pvalue.AsDict(categorical_mapping_dicts),
        )



import json


def load_avro_schema(template_path):
    schema = json.load(open(template_path))

    return schema


def _fill_template_schema(template, var_name, var_values):

    template_fields = template["fields"]
    fields = []

    var_name_in_parentheses = "{" + var_name + "}"

    for f in template_fields:

        if var_name_in_parentheses not in f["name"]:
            fields.append(f)
            continue

        for value in var_values:
            replaced_name = f["name"].format(**{var_name: value})
            fields.append({**f, "name": replaced_name})

    schema = {
        **template,
        "fields": fields,
    }

    return schema


def generate_avro_schema_from_template(template_path, var_name, var_values):
    template = load_avro_schema(template_path)
    return _fill_template_schema(template, var_name, var_values)



from .iter_utils import is_none, is_some, take_as_singleton, take_top, unwrap, unwrap_or_none

__all__ = (
    "is_none",
    "is_some",
    "take_as_singleton",
    "take_top",
    "unwrap",
    "unwrap_or_none",
)



import json

import apache_beam as beam


def Label(**labels):
    """Labels all elements."""
    return "Label" >> beam.Map(lambda r: {**r, **labels})


def Select(*keys):
    """Removes all columns which are not specified in `*keys`."""
    return "Select" >> beam.Map(lambda r: {k: r[k] for k in keys})


def Project(*keys):
    """Transforms each element into a tuple of values of the specified columns."""
    return "Project" >> beam.Map(lambda r: tuple(r[k] for k in keys))


def IndexBy(*keys):
    """Transforms each element `V` into a tuple `(K, V)`.

    `K` is the projection of `V` by `*keys`, which is equal to the tuple
    produced by the `Project` transform.
    """
    return "IndexBy" >> beam.Map(lambda r: (tuple(r[k] for k in keys), r))


def _decimal_default_proc(obj):
    from decimal import Decimal

    if isinstance(obj, Decimal):
        return float(obj)
    raise TypeError


def Stringify():
    """Transforms each element into its JSON representation."""

    def s(obj):
        return json.dumps(obj, default=_decimal_default_proc)

    return "Stringify" >> beam.Map(s)


def IndexBySingle(key):
    """Transforms each element `V` into a tuple `(K, V)`.

    The difference between `IndexBySingle(key)` and `IndexBy(key)` with a single
    key is as follows:
    - `IndexBySingle` produces the index as a plain value.
    - `IndexBy` produces the index as a single-element tuple.
    """
    return "IndexBySingle" >> beam.Map(lambda r: (r[key], r))


def RenameFromTo(from_to_key_mapping):
    """Rename columns according to `from_to_key_mapping`."""

    def rename(row):
        res = dict(row)
        for k1, k2 in from_to_key_mapping.items():
            if k1 in res:
                v = res.pop(k1)
                res[k2] = v
        return res

    return "Rename" >> beam.Map(rename)


def Exclude(*keys):
    """Removes all columns specified in `*keys`."""

    def exclude(row):
        res = dict(row)
        for k in keys:
            if k in res:
                del res[k]
        return res

    return "Exclude" >> beam.Map(exclude)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['carling', 'carling.io', 'carling.iter_utils', 'carling.test_utils']

package_data = \
{'': ['*']}

install_requires = \
['apache-beam>=2.37.0,<3.0.0', 'deepdiff>=5.7.0,<6.0.0']

setup_kwargs = {
    'name': 'carling',
    'version': '0.3.4',
    'description': 'Useful transforms for supporting apache beam pipelines.',
    'long_description': '# Carling\n\n[![CI](https://github.com/mc-digital/carling/actions/workflows/ci.yml/badge.svg)](https://github.com/mc-digital/carling/actions?query=workflow%3ACI)\n[![versions](https://img.shields.io/pypi/pyversions/carling.svg)](https://pypi.org/project/carling/)\n[![pypi](https://img.shields.io/pypi/v/carling)](https://pypi.org/project/carling/)\n[![license](https://img.shields.io/pypi/l/carling)](https://github.com/mc-digital/carling/blob/main/LICENSE)\n\nVia [Wikipedia](<https://en.wikipedia.org/wiki/Carling_(sailing)>):\n\n> Carlings are pieces of timber laid fore and aft under the deck of a ship, from one beam to another.\n> They serve as a foundation for the whole body of the ship.\n\nUseful transforms for supporting our apache beam pipelines.\n\n## Mapping transform utils\n\n#### `carling.Label`\n\nLabels all elements.\n\n#### `carling.Select`\n\nRemoves all columns which are not specified in `*keys`.\n\n#### `carling.Project`\n\nTransforms each element into a tuple of values of the specified columns.\n\n#### `carling.IndexBy`\n\nTransforms each element `V` into a tuple `(K, V)`.\n\n`K` is the projection of `V` by `*keys`, which is equal to the tuple\nproduced by the `Project` transform.\n\n#### `carling.Stringify`\n\nTransforms each element into its JSON representation.\n\n#### `carling.IndexBySingle`\n\nTransforms each element `V` into a tuple `(K, V)`.\n\nThe difference between `IndexBySingle(key)` and `IndexBy(key)` with a single\nkey is as follows:\n\n- `IndexBySingle` produces the index as a plain value.\n- `IndexBy` produces the index as a single-element tuple.\n\n#### `carling.RenameFromTo`\n\nRename columns according to `from_to_key_mapping`.\n\n#### `carling.Exclude`\n\nRemoves all columns specified in `*keys`.\n\n## Grouping transform utils\n\nGeneric grouping transform utils\n\n#### `carling.UniqueOnly`\n\nProduces elements that are the only elements per key after deduplication.\n\nGiven a `PCollection` of `(K, V)`,\nthis transform produces the collection of all `V`s that do not share\nthe same corresponding `K`s with any other elements after deduplicating\nall equivalent `(K, V)` pairs.\n\nThis transform is equivalent to `SingletonOnly` with `apache_beam.Distinct`.\n\n`[(1, "A"), (2, "B1"), (2, "B2"), (3, "C"), (3, "C"), (4, "A")]` will be\ntransformed into `["A", "C", "A"]`.\n\n#### `carling.SingletonOnly`\n\nProduces elements that are the only elements per key.\n\nGiven a `PCollection` of `(K, V)`,\nthis transform produces the collection of all `V`s that do not share\nthe same corresponding `K`s with any other elements.\n\n`[(1, "A"), (2, "B1"), (2, "B2"), (3, "C"), (3, "C"), (4, "A")]` will be\ntransformed into `["A", "A"]`.\n\n#### `carling.Intersection`\n\nProduces the intersection of given `PCollection`s.\n\nGiven a list of `PCollection`s,\nthis transform produces every element that appears in all collections of\nthe list.\nElements are deduplicated before taking the intersection.\n\n#### `carling.FilterByKey`\n\nFilters elements by their keys.\n\nThe constructor receives one or more `PCollection`s of `K`s,\nwhich are regarded as key lists.\nGiven a `PCollection` of `(K, V)`,\nthis transform discards all elements with `K`s that do not appear\nin the key lists.\n\nIf multiple collections are given to the constructor,\nthis transform treats the intersection of them as the key list.\n\n#### `carling.FilterByKeyUsingSideInput`\n\nFilters a single collection by a single lookup collection, using a common key.\n\nGiven: - a `PCollection` (lookup_entries) of `(V)`, as a lookup collection - a `PCollection` (pcoll) of `(V)`, as values to be filtered - a common key (filter_key)\n\nA dictionary called `filter_dict` - is created by mapping the value of `filter_key`\nfor each entry in `lookup_entries` to True.\n\nThen, for each item in pcoll, the value associated with `filter_key` checkd against\n`filter_dict`, and if it is found, the entry passes through. Otherwise, the entry is\ndiscarded.\n\nNote: `lookup_entries` will be used as a **side input**, so care\nmust be taken regarding the size of the `lookup_entries`\n\n#### `carling.DifferencePerKey`\n\nProduces the difference per key between two `PCollection`s.\n\nGiven two `PCollection`s of `V`,\nthis transform indexes the collections by the specified keys `primary_keys`,\ncompares corresponding two `V` lists for every `K`,\nand produces the difference per `K`.\nIf there is no difference, this transform produces nothing.\n\nTwo `V` lists are considered to be different if the numbers of elements\ndiffer or two elements of the lists with a same index differ\nat one of the specified columns `columns`.\n\n#### `carling.SortedSelectPerKey`\n\n- Groups items by a set of `keys` -- column names per row\n- Emits the "MAX" _value_ for each collection as defined by the `key_fn`\n- Can emit "MIN" by passing `reverse=True` kwarg\n\n#### `carling.PartitionRowsContainingNone`\n\nEmits two tagged PCollections:\n\n- Default (`result[None]`): Rows are guaranteed not to have any `None` values\n- `result["contains_none"]`: Rows for which at least one column had a `None` value\n\n## Categorical\n\n#### `carling.CreateCategoricalDicts`\n\nFor a set of columnular data inputs this function takes:\n\n    - cat_cols:\n\n        Type: `[str]`\n\n        An array of "categorical" columns\n\n    - existing_dicts:\n\n        Type: `PCollection[(string, string, int)]`\n\n        Rows of tuples of type:\n        (column, previously_seen_value, mapped_unique_int)\n\n        Mapping a set of "previously seen" keys to unique int values for each\n        column.\n        Not optional.\n        If none exist, pass an empty PCollection.\n\nIt then creates a transform which takes a pcollection and\n\n    - looks at the input pcoll for unseen values in each categorical column\n    - creates new unique integers for each distinct unseen value, starting at max(previous value for column)+1\n    - ammends the existing mappings with (col, unseen_value, new_unique_int)\n\nOutput is:\n\n    - Type: `PCollection[(string, string, int)]`\n\nThis is useful for preparing data to be trained by eg. LightGBM\n\n#### `carling.ReplaceCategoricalColumns`\n\n- Utilizes the "categorical dictionary rows" generated by `CreateCategoricalDicts` which is a list of pairs of type of `(column, value,unique_int)`.\n\n- Replaces each column with the appropriate value found in the mapping.\n\n## Test Utils\n\n#### `carling.test_utils.pprint_equal_to`\n\nThis module contains the `equal_to` function from apache beam, but adapted to output results using pretty print. Reading the results as a large, unformatted string makes it harder to pick out what changed/is missing.\n\n## General Util\n\n#### `carling.LogSample`\n\nPrint items of the given `PCollection` to the log.\n\n`LogSample` prints the JSON representations of the input items to the Python\'s\nstandard logging system.\n\nTo avoid too much log entries being printed, `LogSample` limits the number of\nlogged items. The constructor parameter `n` determines the limit.\n\nBy default, `LogSample` prints logs with the `INFO` log level. The constructor\nparameter `level` determines the level.\n\n#### `carling.ReifyMultiValueOption`\n\nPrepares multi-value delimited options. Useful in contexts where\nyou want to create a multi-value option in a template environment.\n\n- inputs:\n  - delimited string option\n  - optional delimiter string (default is "|")\n\n* output:\n  - Type: `PCollection[str]`\n',
    'author': 'Adam Moore',
    'author_email': 'adam@mcdigital.jp',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mc-digital/carling',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
